@@ -111,9 +111,10 @@ def annotate_file(branch, rev_id, file_id, verbose=False, full=False,
 
 
 def _annotations(repo, file_id, rev_id):
-    """Return the list of (origin,text) for a revision of a file in a repository."""
-    w = repo.weave_store.get_weave(file_id, repo.get_transaction())
-    return w.annotate(rev_id)
+    """Return the list of (origin_revision_id, line_text) for a revision of a file in a repository."""
+    annotations = repo.texts.annotate((file_id, rev_id))
+    # 
+    return [(key[-1], line) for (key, line) in annotations]
 
 
 def _annotate_file(branch, rev_id, file_id):
@@ -284,16 +285,9 @@ def _find_matching_unannotated_lines(output_lines, plain_child_lines,
                     if len(heads) == 1:
                         output_append((iter(heads).next(), left[1]))
                     else:
-                        # Both claim different origins
-                        output_append((revision_id, left[1]))
-                        # We know that revision_id is the head for
-                        # left and right, so cache it
-                        heads_provider.cache(
-                            (revision_id, left[0]),
-                            (revision_id,))
-                        heads_provider.cache(
-                            (revision_id, right[0]),
-                            (revision_id,))
+                        # Both claim different origins, sort lexicographically
+                        # so that we always get a stable result.
+                        output_append(sorted([left, right])[0])
         last_child_idx = child_idx + match_len
 
 
