@@ -55,9 +55,9 @@ def zip_exporter(tree, dest, root):
         entries = inv.iter_entries()
         entries.next() # skip root
         for dp, ie in entries:
-            # .bzrignore has no meaning outside of a working tree
-            # so do not export it
-            if dp == ".bzrignore":
+            # The .bzr* namespace is reserved for "magic" files like
+            # .bzrignore and .bzrrules - do not export these
+            if dp.startswith(".bzr"):
                 continue
 
             file_id = ie.file_id
